@@ -16,36 +16,36 @@
 #include <nb30.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <locale.h>
+#include <mbctype.h>
 #include <winsock2.h>
+#include <ErrorRep.h>
 
 #include <osi.h>
 #include "afsd.h"
 #include <rx\rx.h>
 #include <rx\rx_null.h>
 #include <WINNT/syscfg.h>
+#include <WINNT/afsreg.h>
 
 #include "smb.h"
 #include "cm_rpc.h"
 #include "lanahelper.h"
 #include <strsafe.h>
 #include "afsicf.h"
+#include "cm_memmap.h"
 
 extern int RXAFSCB_ExecuteRequest(struct rx_call *z_call);
 extern int RXSTATS_ExecuteRequest(struct rx_call *z_call);
 
 extern afs_int32 cryptall;
 
-char AFSConfigKeyName[] =
-	"SYSTEM\\CurrentControlSet\\Services\\TransarcAFSDaemon\\Parameters";
-
 osi_log_t *afsd_logp;
 
-char cm_rootVolumeName[64];
+cm_config_data_t        cm_data;
+
+char cm_rootVolumeName[VL_MAXNAMELEN];
 DWORD cm_rootVolumeNameLen;
-cm_volume_t *cm_rootVolumep = NULL;
-cm_cell_t *cm_rootCellp = NULL;
-cm_fid_t cm_rootFid;
-cm_scache_t *cm_rootSCachep = NULL;
 char cm_mountRoot[1024];
 DWORD cm_mountRootLen;
 int cm_logChunkSize;
@@ -54,13 +54,13 @@ int cm_chunkSize;
 char *cm_FakeRootDir;
 #endif /* freelance */
 
-int smb_UseV3;
+int smb_UseV3 = 1;
 
 int LANadapter;
 
 int numBkgD;
 int numSvThreads;
-
+long rx_mtu = -1;
 int traceOnPanic = 0;
 
 int logReady = 0;
@@ -72,6 +72,7 @@ char cm_NetbiosName[MAX_NB_NAME_LENGTH] = "";
 
 char cm_CachePath[MAX_PATH];
 DWORD cm_CachePathLen;
+DWORD cm_ValidateCache = 1;
 
 BOOL isGateway = FALSE;
 
@@ -99,85 +100,22 @@ HANDLE afsi_file;
 int cm_dnsEnabled = 1;
 #endif
 
-extern initUpperCaseTable();
-void afsd_initUpperCaseTable() 
-{
-	initUpperCaseTable();
-}
-
-void
-afsi_start()
-{
-	char wd[100];
-	char t[100], u[100], *p, *path;
-	int zilch;
-	int code;
-    DWORD dwLow, dwHigh;
-	HKEY parmKey;
-	DWORD dummyLen;
-    DWORD maxLogSize = 100 * 1024;
-
-	afsi_file = INVALID_HANDLE_VALUE;
-    if (getenv("TEMP"))
-    {
-        StringCbCopyA(wd, sizeof(wd), getenv("TEMP"));
-    }
-    else
-    {
-        code = GetWindowsDirectory(wd, sizeof(wd));
-        if (code == 0) return;
-    }
-	StringCbCatA(wd, sizeof(wd), "\\afsd_init.log");
-	GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, t, sizeof(t));
-	afsi_file = CreateFile(wd, GENERIC_WRITE, FILE_SHARE_READ, NULL,
-                           OPEN_ALWAYS, FILE_FLAG_WRITE_THROUGH, NULL);
-
-    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSConfigKeyName,
-                         0, KEY_QUERY_VALUE, &parmKey);
-	if (code == ERROR_SUCCESS) {
-        dummyLen = sizeof(maxLogSize);
-        code = RegQueryValueEx(parmKey, "MaxLogSize", NULL, NULL,
-                                (BYTE *) &maxLogSize, &dummyLen);
-        RegCloseKey (parmKey);
-	}
-
-    if (maxLogSize) {
-        dwLow = GetFileSize( afsi_file, &dwHigh );
-        if ( dwHigh > 0 || dwLow >= maxLogSize ) {
-            CloseHandle(afsi_file);
-            afsi_file = CreateFile( wd, GENERIC_WRITE, FILE_SHARE_READ, NULL,
-                                    CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, NULL);
-        }
-    }
-
-    SetFilePointer(afsi_file, 0, NULL, FILE_END);
-	GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, u, sizeof(u));
-	StringCbCatA(t, sizeof(t), ": Create log file\n");
-	StringCbCatA(u, sizeof(u), ": Created log file\n");
-	WriteFile(afsi_file, t, strlen(t), &zilch, NULL);
-	WriteFile(afsi_file, u, strlen(u), &zilch, NULL);
-    p = "PATH=";
-    path = getenv("PATH");
-	WriteFile(afsi_file, p, strlen(p), &zilch, NULL);
-	WriteFile(afsi_file, path, strlen(path), &zilch, NULL);
-	WriteFile(afsi_file, "\n", 1, &zilch, NULL);
-}
 
 static int afsi_log_useTimestamp = 1;
 
 void
 afsi_log(char *pattern, ...)
 {
-	char s[256], t[100], d[100], u[512];
-	int zilch;
-	va_list ap;
-	va_start(ap, pattern);
+    char s[256], t[100], d[100], u[512];
+    DWORD zilch;
+    va_list ap;
+    va_start(ap, pattern);
 
-	StringCbVPrintfA(s, sizeof(s), pattern, ap);
+    StringCbVPrintfA(s, sizeof(s), pattern, ap);
     if ( afsi_log_useTimestamp ) {
         GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, t, sizeof(t));
-		GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, d, sizeof(d));
-		StringCbPrintfA(u, sizeof(u), "%s %s: %s\n", d, t, s);
+        GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, d, sizeof(d));
+        StringCbPrintfA(u, sizeof(u), "%s %s: %s\n", d, t, s);
         if (afsi_file != INVALID_HANDLE_VALUE)
             WriteFile(afsi_file, u, strlen(u), &zilch, NULL);
 #ifdef NOTSERVICE
@@ -189,31 +127,110 @@ afsi_log(char *pattern, ...)
     }
 }
 
+extern initUpperCaseTable();
+void afsd_initUpperCaseTable() 
+{
+    initUpperCaseTable();
+}
+
+void
+afsi_start()
+{
+    char wd[256];
+    char t[100], u[100], *p, *path;
+    int zilch;
+    int code;
+    DWORD dwLow, dwHigh;
+    HKEY parmKey;
+    DWORD dummyLen;
+    DWORD maxLogSize = 100 * 1024;
+
+    afsi_file = INVALID_HANDLE_VALUE;
+    code = GetEnvironmentVariable("TEMP", wd, sizeof(wd));
+    if ( code == 0 || code > sizeof(wd) )
+    {
+        code = GetWindowsDirectory(wd, sizeof(wd));
+        if (code == 0) 
+            return;
+    }
+    StringCbCatA(wd, sizeof(wd), "\\afsd_init.log");
+    GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, t, sizeof(t));
+    afsi_file = CreateFile(wd, GENERIC_WRITE, FILE_SHARE_READ, NULL,
+                           OPEN_ALWAYS, FILE_FLAG_WRITE_THROUGH, NULL);
+
+    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSREG_CLT_SVC_PARAM_SUBKEY,
+                         0, KEY_QUERY_VALUE, &parmKey);
+    if (code == ERROR_SUCCESS) {
+        dummyLen = sizeof(maxLogSize);
+        code = RegQueryValueEx(parmKey, "MaxLogSize", NULL, NULL,
+                                (BYTE *) &maxLogSize, &dummyLen);
+        RegCloseKey (parmKey);
+    }
+
+    if (maxLogSize) {
+        dwLow = GetFileSize( afsi_file, &dwHigh );
+        if ( dwHigh > 0 || dwLow >= maxLogSize ) {
+            CloseHandle(afsi_file);
+            afsi_file = CreateFile( wd, GENERIC_WRITE, FILE_SHARE_READ, NULL,
+                                    CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, NULL);
+        }
+    }
+
+    SetFilePointer(afsi_file, 0, NULL, FILE_END);
+    GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, u, sizeof(u));
+    StringCbCatA(t, sizeof(t), ": Create log file\n");
+    StringCbCatA(u, sizeof(u), ": Created log file\n");
+    WriteFile(afsi_file, t, strlen(t), &zilch, NULL);
+    WriteFile(afsi_file, u, strlen(u), &zilch, NULL);
+    p = "PATH=";
+    code = GetEnvironmentVariable("PATH", NULL, 0);
+    path = malloc(code);
+    code = GetEnvironmentVariable("PATH", path, code);
+    WriteFile(afsi_file, p, strlen(p), &zilch, NULL);
+    WriteFile(afsi_file, path, strlen(path), &zilch, NULL);
+    WriteFile(afsi_file, "\n", 1, &zilch, NULL);
+    free(path);
+
+    /* Initialize C RTL Code Page conversion functions */
+    /* All of the path info obtained from the SMB client is in the OEM code page */
+    afsi_log("OEM Code Page = %d", GetOEMCP());
+    afsi_log("locale =  %s", setlocale(LC_ALL,NULL));
+#ifdef COMMENT
+    /* Two things to look into.  First, should mbstowcs() be performing 
+     * character set translations from OEM to Unicode in smb3.c; 
+     * Second, do we need to set this translation in each function 
+     * due to multi-threading. 
+     */
+    afsi_log("locale -> %s", setlocale(LC_ALL, ".OCP"));
+    afsi_log("_setmbcp = %d -> %d", _setmbcp(_MB_CP_OEM), _getmbcp());
+#endif /* COMMENT */
+}
+
 /*
  * Standard AFSD trace
  */
 
 void afsd_ForceTrace(BOOL flush)
 {
-	HANDLE handle;
-	int len;
-	char buf[256];
+    HANDLE handle;
+    int len;
+    char buf[256];
 
-	if (!logReady) 
+    if (!logReady) 
         return;
 
-	len = GetTempPath(sizeof(buf)-10, buf);
-	StringCbCopyA(&buf[len], sizeof(buf)-len, "/afsd.log");
-	handle = CreateFile(buf, GENERIC_WRITE, FILE_SHARE_READ,
-			    NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
-	if (handle == INVALID_HANDLE_VALUE) {
-		logReady = 0;
-		osi_panic("Cannot create log file", __FILE__, __LINE__);
-	}
-	osi_LogPrint(afsd_logp, handle);
-	if (flush)
-		FlushFileBuffers(handle);
-	CloseHandle(handle);
+    len = GetTempPath(sizeof(buf)-10, buf);
+    StringCbCopyA(&buf[len], sizeof(buf)-len, "/afsd.log");
+    handle = CreateFile(buf, GENERIC_WRITE, FILE_SHARE_READ,
+                         NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
+    if (handle == INVALID_HANDLE_VALUE) {
+        logReady = 0;
+        osi_panic("Cannot create log file", __FILE__, __LINE__);
+    }
+    osi_LogPrint(afsd_logp, handle);
+    if (flush)
+        FlushFileBuffers(handle);
+    CloseHandle(handle);
 }
 
 static void
@@ -243,6 +260,10 @@ configureBackConnectionHostNames(void)
      *             check for the UnsetDisableLoopbackCheck value.  
      *             If set, set the DisableLoopbackCheck flag to 0x0 
      *             and delete the UnsetDisableLoopbackCheck value
+     *
+     * Starting in Longhorn Beta 1, an entry in the BackConnectionHostNames value will
+     * force Windows to use the loopback authentication mechanism for the specified 
+     * services.
      */
     HKEY hkLsa;
     HKEY hkMSV10;
@@ -250,7 +271,7 @@ configureBackConnectionHostNames(void)
     DWORD dwType;
     DWORD dwSize;
     DWORD dwValue;
-    PBYTE pHostNames = NULL, pName;
+    PBYTE pHostNames = NULL, pName = NULL;
     BOOL  bNameFound = FALSE;   
 
     if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
@@ -273,12 +294,14 @@ configureBackConnectionHostNames(void)
         }
              
         if ( !bNameFound ) {
+            int size = strlen(cm_NetbiosName) + 2;
             if ( !pHostNames ) {
-                pName = pHostNames = malloc(strlen(cm_NetbiosName) + 2);
+                pHostNames = malloc(size);
                 dwSize = 1;
             }
-            strcpy(pName, cm_NetbiosName);
-            pName += strlen(cm_NetbiosName) + 1;
+            pName = pHostNames;
+            StringCbCopyA(pName, size, cm_NetbiosName);
+            pName += size - 1;
             *pName = '\0';  /* add a second nul terminator */
 
             dwType = REG_MULTI_SZ;
@@ -292,15 +315,15 @@ configureBackConnectionHostNames(void)
                                &hkLsa) == ERROR_SUCCESS )
             {
                 dwSize = sizeof(DWORD);
-                if ( RegQueryValueEx( hkLsa, "DisableLoopbackCheck", 0, &dwType, &dwValue, &dwSize) != ERROR_SUCCESS ||
+                if ( RegQueryValueEx( hkLsa, "DisableLoopbackCheck", 0, &dwType, (LPBYTE)&dwValue, &dwSize) != ERROR_SUCCESS ||
                      dwValue == 0 ) {
                     dwType = REG_DWORD;
                     dwSize = sizeof(DWORD);
                     dwValue = 1;
-                    RegSetValueEx( hkLsa, "DisableLoopbackCheck", 0, dwType, &dwValue, dwSize);
+                    RegSetValueEx( hkLsa, "DisableLoopbackCheck", 0, dwType, (LPBYTE)&dwValue, dwSize);
 
                     if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, 
-                                        "SOFTWARE\\OpenAFS\\Client",
+                                        AFSREG_CLT_OPENAFS_SUBKEY,
                                         0,
                                         NULL,
                                         REG_OPTION_NON_VOLATILE,
@@ -312,7 +335,7 @@ configureBackConnectionHostNames(void)
                         dwType = REG_DWORD;
                         dwSize = sizeof(DWORD);
                         dwValue = 1;
-                        RegSetValueEx( hkClient, "RemoveDisableLoopbackCheck", 0, dwType, &dwValue, dwSize);
+                        RegSetValueEx( hkClient, "RemoveDisableLoopbackCheck", 0, dwType, (LPBYTE)&dwValue, dwSize);
                         RegCloseKey(hkClient);
                     }
                     RegCloseKey(hkLsa);
@@ -320,7 +343,7 @@ configureBackConnectionHostNames(void)
             }
         } else {
             if (RegCreateKeyEx( HKEY_LOCAL_MACHINE, 
-                                "SOFTWARE\\OpenAFS\\Client",
+                                AFSREG_CLT_OPENAFS_SUBKEY,
                                 0,
                                 NULL,
                                 REG_OPTION_NON_VOLATILE,
@@ -330,7 +353,7 @@ configureBackConnectionHostNames(void)
                                 NULL) == ERROR_SUCCESS) {
 
                 dwSize = sizeof(DWORD);
-                if ( RegQueryValueEx( hkClient, "RemoveDisableLoopbackCheck", 0, &dwType, &dwValue, &dwSize) == ERROR_SUCCESS &&
+                if ( RegQueryValueEx( hkClient, "RemoveDisableLoopbackCheck", 0, &dwType, (LPBYTE)&dwValue, &dwSize) == ERROR_SUCCESS &&
                      dwValue == 1 ) {
                     if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                                        "SYSTEM\\CurrentControlSet\\Control\\Lsa",
@@ -350,6 +373,157 @@ configureBackConnectionHostNames(void)
     }
 }
 
+#if !defined(DJGPP)
+static void afsd_InitServerPreferences(void)
+{
+    HKEY hkPrefs = 0;
+    DWORD dwType, dwSize;
+    DWORD dwPrefs = 0;
+    DWORD dwIndex;
+    TCHAR szHost[256];
+    DWORD dwHostSize = 256;
+    DWORD dwRank;
+    struct sockaddr_in	saddr;
+    cm_server_t       *tsp;
+
+    if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
+                      AFSREG_CLT_OPENAFS_SUBKEY "\\Server Preferences\\VLDB",
+                      0,
+                      KEY_READ|KEY_QUERY_VALUE,
+                      &hkPrefs) == ERROR_SUCCESS) {
+
+        RegQueryInfoKey( hkPrefs,
+                         NULL,  /* lpClass */
+                         NULL,  /* lpcClass */
+                         NULL,  /* lpReserved */
+                         NULL,  /* lpcSubKeys */
+                         NULL,  /* lpcMaxSubKeyLen */
+                         NULL,  /* lpcMaxClassLen */
+                         &dwPrefs, /* lpcValues */
+                         NULL,  /* lpcMaxValueNameLen */
+                         NULL,  /* lpcMaxValueLen */
+                         NULL,  /* lpcbSecurityDescriptor */
+                         NULL   /* lpftLastWriteTime */
+                         );
+
+        for ( dwIndex = 0 ; dwIndex < dwPrefs; dwIndex++ ) {
+
+            dwSize = sizeof(DWORD);
+            dwHostSize = 256;
+
+            if (RegEnumValue( hkPrefs, dwIndex, szHost, &dwHostSize, NULL,
+                              &dwType, (LPBYTE)&dwRank, &dwSize))
+            {
+                afsi_log("RegEnumValue(hkPrefs) failed");
+                continue;
+            }
+
+            afsi_log("VLDB Server Preference: %s = %d",szHost, dwRank);
+
+            if (isdigit(szHost[0]))
+            {
+                if ((saddr.sin_addr.S_un.S_addr = inet_addr (szHost)) == INADDR_NONE)
+                    continue;
+            } else {
+                HOSTENT *pEntry;
+                if ((pEntry = gethostbyname (szHost)) == NULL)
+                    continue;
+
+                saddr.sin_addr.S_un.S_addr = *(unsigned long *)pEntry->h_addr;
+            }
+            saddr.sin_family = AF_INET;
+            dwRank += (rand() & 0x000f);
+
+            tsp = cm_FindServer(&saddr, CM_SERVER_VLDB);
+            if ( tsp )		/* an existing server - ref count increased */
+            {
+                tsp->ipRank = (USHORT)dwRank; /* no need to protect by mutex*/
+
+                /* set preferences for an existing vlserver */
+                cm_ChangeRankCellVLServer(tsp);
+                cm_PutServer(tsp);  /* decrease refcount */
+            }
+            else	/* add a new server without a cell */
+            {
+                tsp = cm_NewServer(&saddr, CM_SERVER_VLDB, NULL); /* refcount = 1 */
+                tsp->ipRank = (USHORT)dwRank;
+            }
+        }
+
+        RegCloseKey(hkPrefs);
+    }
+
+    if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
+                      AFSREG_CLT_OPENAFS_SUBKEY "\\Server Preferences\\File",
+                      0,
+                      KEY_READ|KEY_QUERY_VALUE,
+                      &hkPrefs) == ERROR_SUCCESS) {
+
+        RegQueryInfoKey( hkPrefs,
+                         NULL,  /* lpClass */
+                         NULL,  /* lpcClass */
+                         NULL,  /* lpReserved */
+                         NULL,  /* lpcSubKeys */
+                         NULL,  /* lpcMaxSubKeyLen */
+                         NULL,  /* lpcMaxClassLen */
+                         &dwPrefs, /* lpcValues */
+                         NULL,  /* lpcMaxValueNameLen */
+                         NULL,  /* lpcMaxValueLen */
+                         NULL,  /* lpcbSecurityDescriptor */
+                         NULL   /* lpftLastWriteTime */
+                         );
+
+        for ( dwIndex = 0 ; dwIndex < dwPrefs; dwIndex++ ) {
+
+            dwSize = sizeof(DWORD);
+            dwHostSize = 256;
+
+            if (RegEnumValue( hkPrefs, dwIndex, szHost, &dwHostSize, NULL,
+                              &dwType, (LPBYTE)&dwRank, &dwSize))
+            {
+                afsi_log("RegEnumValue(hkPrefs) failed");
+                continue;
+            }
+
+            afsi_log("File Server Preference: %s = %d",szHost, dwRank);
+
+            if (isdigit(szHost[0]))
+            {
+                if ((saddr.sin_addr.S_un.S_addr = inet_addr (szHost)) == INADDR_NONE)
+                    continue;
+            } else {
+                HOSTENT *pEntry;
+                if ((pEntry = gethostbyname (szHost)) == NULL)
+                    continue;
+
+                saddr.sin_addr.S_un.S_addr = *(unsigned long *)pEntry->h_addr;
+            }
+            saddr.sin_family = AF_INET;
+            dwRank += (rand() & 0x000f);
+
+            tsp = cm_FindServer(&saddr, CM_SERVER_FILE);
+            if ( tsp )		/* an existing server - ref count increased */
+            {
+                tsp->ipRank = (USHORT)dwRank; /* no need to protect by mutex*/
+
+                /* find volumes which might have RO copy 
+                /* on server and change the ordering of 
+                 * their RO list 
+                 */
+                cm_ChangeRankVolume(tsp);
+                cm_PutServer(tsp);  /* decrease refcount */
+            }
+            else	/* add a new server without a cell */
+            {
+                tsp = cm_NewServer(&saddr, CM_SERVER_FILE, NULL); /* refcount = 1 */
+                tsp->ipRank = (USHORT)dwRank;
+            }
+        }
+
+        RegCloseKey(hkPrefs);
+    }
+}
+#endif /* DJGPP */
 
 /*
  * AFSD Initialization
@@ -358,15 +532,15 @@ configureBackConnectionHostNames(void)
 int afsd_InitCM(char **reasonP)
 {
     osi_uid_t debugID;
-    long cacheBlocks;
-    long cacheSize;
+    DWORD cacheBlocks;
+    DWORD cacheSize;
     long logChunkSize;
-    long stats;
+    DWORD stats;
     long traceBufSize;
     long maxcpus;
     long ltt, ltto;
-    long rx_mtu, rx_nojumbo;
-    long virtualCache;
+    long rx_nojumbo;
+    long virtualCache = 0;
     char rootCellName[256];
     struct rx_service *serverp;
     static struct rx_securityClass *nullServerSecurityClassp;
@@ -381,6 +555,12 @@ int afsd_InitCM(char **reasonP)
     WSADATA WSAjunk;
     lana_number_t lanaNum;
     int i;
+    char *p, *q; 
+    int cm_noIPAddr;         /* number of client network interfaces */
+    int cm_IPAddr[CM_MAXINTERFACE_ADDR];    /* client's IP address in host order */
+    int cm_SubnetMask[CM_MAXINTERFACE_ADDR];/* client's subnet mask in host order*/
+    int cm_NetMtu[CM_MAXINTERFACE_ADDR];    /* client's MTU sizes */
+    int cm_NetFlags[CM_MAXINTERFACE_ADDR];  /* network flags */
 
     WSAStartup(0x0101, &WSAjunk);
 
@@ -407,7 +587,7 @@ int afsd_InitCM(char **reasonP)
     srand(ntohl(cm_HostAddr));
 
     /* Look up configuration parameters in Registry */
-    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSConfigKeyName,
+    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSREG_CLT_SVC_PARAM_SUBKEY,
                          0, KEY_QUERY_VALUE, &parmKey);
     if (code != ERROR_SUCCESS) {
         FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM
@@ -591,7 +771,7 @@ int afsd_InitCM(char **reasonP)
     if (code == ERROR_SUCCESS && buf[0]) {
         if (regType == REG_EXPAND_SZ) {
             dummyLen = ExpandEnvironmentStrings(buf, cm_CachePath, sizeof(cm_CachePath));
-            if(dummyLen > sizeof(cm_CachePath)) {
+            if (dummyLen > sizeof(cm_CachePath)) {
                 afsi_log("Cache path [%s] longer than %d after expanding env strings", buf, sizeof(cm_CachePath));
                 osi_panic("CachePath too long", __FILE__, __LINE__);
             }
@@ -600,32 +780,45 @@ int afsd_InitCM(char **reasonP)
         }
         afsi_log("Cache path %s", cm_CachePath);
     } else {
-        GetWindowsDirectory(cm_CachePath, sizeof(cm_CachePath));
-        cm_CachePath[2] = 0;	/* get drive letter only */
-        StringCbCatA(cm_CachePath, sizeof(cm_CachePath), "\\AFSCache");
+        dummyLen = ExpandEnvironmentStrings("%TEMP%\\AFSCache", cm_CachePath, sizeof(cm_CachePath));
+        if (dummyLen > sizeof(cm_CachePath)) {
+            afsi_log("Cache path [%%TEMP%%\\AFSCache] longer than %d after expanding env strings", 
+                     sizeof(cm_CachePath));
+            osi_panic("CachePath too long", __FILE__, __LINE__);
+        }
         afsi_log("Default cache path %s", cm_CachePath);
     }
 
     dummyLen = sizeof(virtualCache);
     code = RegQueryValueEx(parmKey, "NonPersistentCaching", NULL, NULL,
-                            &virtualCache, &dummyLen);
-    if (code == ERROR_SUCCESS && virtualCache) {
-        buf_cacheType = CM_BUF_CACHETYPE_VIRTUAL;
-    } else {
-        buf_cacheType = CM_BUF_CACHETYPE_FILE;
+                            (LPBYTE)&virtualCache, &dummyLen);
+    afsi_log("Cache type is %s", (virtualCache?"VIRTUAL":"FILE"));
+
+    if (!virtualCache) {
+        dummyLen = sizeof(cm_ValidateCache);
+        code = RegQueryValueEx(parmKey, "ValidateCache", NULL, NULL,
+                               (LPBYTE)&cm_ValidateCache, &dummyLen);
+        if ( cm_ValidateCache < 0 || cm_ValidateCache > 2 )
+            cm_ValidateCache = 1;
+        switch (cm_ValidateCache) {
+        case 0:
+            afsi_log("Cache Validation disabled");
+            break;
+        case 1:
+            afsi_log("Cache Validation on Startup");
+            break;
+        case 2:
+            afsi_log("Cache Validation on Startup and Shutdown");
+            break;
+        }
     }
-    afsi_log("Cache type is %s", ((buf_cacheType == CM_BUF_CACHETYPE_FILE)?"FILE":"VIRTUAL"));
 
     dummyLen = sizeof(traceOnPanic);
     code = RegQueryValueEx(parmKey, "TrapOnPanic", NULL, NULL,
                             (BYTE *) &traceOnPanic, &dummyLen);
-    if (code == ERROR_SUCCESS)
-        afsi_log("Set to %s on panic",
-                  traceOnPanic ? "trap" : "not trap");
-    else {  
-        traceOnPanic = 0;
-        /* Don't log */
-    }
+    if (code != ERROR_SUCCESS)
+        traceOnPanic = 1;              /* log */
+    afsi_log("Set to %s on panic", traceOnPanic ? "trap" : "not trap");
 
     dummyLen = sizeof(reportSessionStartups);
     code = RegQueryValueEx(parmKey, "ReportSessionStartups", NULL, NULL,
@@ -646,35 +839,36 @@ int afsd_InitCM(char **reasonP)
 
     dummyLen = sizeof(buf);
     code = RegQueryValueEx(parmKey, "SysName", NULL, NULL, buf, &dummyLen);
-    if (code == ERROR_SUCCESS && buf[0]) {
-        char * p, *q; 
-        afsi_log("Sys name %s", buf);
-
-        for (p = q = buf; p < buf + dummyLen; p++)
-        {
-            if (*p == '\0' || isspace(*p)) {
-                memcpy(cm_sysNameList[cm_sysNameCount],q,p-q);
-                cm_sysNameList[cm_sysNameCount][p-q] = '\0';
-                cm_sysNameCount++;
-
-                do {
-                    if (*p == '\0')
-                        goto done_sysname;
-                        
-                    p++;
-                } while (*p == '\0' || isspace(*p));
-                q = p;
-                p--;
-            }
-        }
-      done_sysname:
-        StringCbCopyA(cm_sysName, MAXSYSNAME, cm_sysNameList[0]);
-    } else {
-        cm_sysNameCount = 1;
-        StringCbCopyA(cm_sysName, MAXSYSNAME, "i386_nt40");
-        StringCbCopyA(cm_sysNameList[0], MAXSYSNAME, "i386_nt40");
-        afsi_log("Default sys name %s", cm_sysName);
+    if (code != ERROR_SUCCESS || !buf[0]) {
+#if defined(_IA64_)
+        StringCbCopyA(buf, sizeof(buf), "ia64_win64");
+#elif defined(_AMD64)
+        StringCbCopyA(buf, sizeof(buf), "amd64_win64");
+#else /* assume x86 32-bit */
+        StringCbCopyA(buf, sizeof(buf), "x86_win32 i386_w2k i386_nt40");
+#endif
     }
+    afsi_log("Sys name %s", buf); 
+
+    /* breakup buf into individual search string entries */
+    for (p = q = buf; p < buf + dummyLen; p++)
+    {
+        if (*p == '\0' || isspace(*p)) {
+            memcpy(cm_sysNameList[cm_sysNameCount],q,p-q);
+            cm_sysNameList[cm_sysNameCount][p-q] = '\0';
+            cm_sysNameCount++;
+
+            do {
+                if (*p == '\0')
+                    goto done_sysname;
+                p++;
+            } while (*p == '\0' || isspace(*p));
+            q = p;
+            p--;
+        }
+    }
+  done_sysname:
+    StringCbCopyA(cm_sysName, MAXSYSNAME, cm_sysNameList[0]);
 
     dummyLen = sizeof(cryptall);
     code = RegQueryValueEx(parmKey, "SecurityLevel", NULL, NULL,
@@ -800,7 +994,7 @@ int afsd_InitCM(char **reasonP)
     RegCloseKey (parmKey);
 
     /* Call lanahelper to get Netbios name, lan adapter number and gateway flag */
-    if(SUCCEEDED(code = lana_GetUncServerNameEx(cm_NetbiosName, &lanaNum, &isGateway, LANA_NETBIOS_NAME_FULL))) {
+    if (SUCCEEDED(code = lana_GetUncServerNameEx(cm_NetbiosName, &lanaNum, &isGateway, LANA_NETBIOS_NAME_FULL))) {
         LANadapter = (lanaNum == LANA_INVALID)? -1: lanaNum;
 
         if (LANadapter != -1)
@@ -818,13 +1012,7 @@ int afsd_InitCM(char **reasonP)
         osi_panic(buf, __FILE__, __LINE__);
     }
 
-    /* setup early variables */
-    /* These both used to be configurable. */
-    smb_UseV3 = 1;
-    buf_bufferSize = CM_CONFIGDEFAULT_BLOCKSIZE;
-
-    /* turn from 1024 byte units into memory blocks */
-    cacheBlocks = (cacheSize * 1024) / buf_bufferSize;
+    cacheBlocks = (cacheSize * 1024) / CM_CONFIGDEFAULT_BLOCKSIZE;
         
     /* get network related info */
     cm_noIPAddr = CM_MAXINTERFACE_ADDR;
@@ -843,13 +1031,13 @@ int afsd_InitCM(char **reasonP)
      */
     cm_initParams.nChunkFiles = 0;
     cm_initParams.nStatCaches = stats;
-    cm_initParams.nDataCaches = 0;
-    cm_initParams.nVolumeCaches = 0;
+    cm_initParams.nDataCaches = cacheBlocks;
+    cm_initParams.nVolumeCaches = stats/2;
     cm_initParams.firstChunkSize = cm_chunkSize;
     cm_initParams.otherChunkSize = cm_chunkSize;
     cm_initParams.cacheSize = cacheSize;
     cm_initParams.setTime = 0;
-    cm_initParams.memCache = 0;
+    cm_initParams.memCache = 1;
 
     /* Set RX parameters before initializing RX */
     if ( rx_nojumbo ) {
@@ -868,6 +1056,34 @@ int afsd_InitCM(char **reasonP)
     /* Ensure the AFS Netbios Name is registered to allow loopback access */
     configureBackConnectionHostNames();
 
+    /* init user daemon, and other packages */
+    cm_InitUser();
+
+    cm_InitConn();
+
+    cm_InitServer();
+        
+    cm_InitIoctl();
+        
+    smb_InitIoctl();
+        
+    cm_InitCallback();
+        
+    code = cm_InitMappedMemory(virtualCache, cm_CachePath, stats, cm_chunkSize, cacheBlocks);
+    afsi_log("cm_InitMappedMemory code %x", code);
+    if (code != 0) {
+        *reasonP = "error initializing cache file";
+        return -1;
+    }
+
+#ifdef AFS_AFSDB_ENV
+#if !defined(_WIN32_WINNT) || (_WIN32_WINNT < 0x0500)
+    if (cm_InitDNS(cm_dnsEnabled) == -1)
+        cm_dnsEnabled = 0;  /* init failed, so deactivate */
+    afsi_log("cm_InitDNS %d", cm_dnsEnabled);
+#endif
+#endif
+
     /* initialize RX, and tell it to listen to port 7001, which is used for
      * callback RPC messages.
      */
@@ -878,9 +1094,6 @@ int afsd_InitCM(char **reasonP)
         return -1;
     }
 
-    /* Initialize the RPC server for session keys */
-    RpcInit();
-
     /* create an unauthenticated service #1 for callbacks */
     nullServerSecurityClassp = rxnull_NewServerSecurityObject();
     serverp = rx_NewService(0, 1, "AFS", &nullServerSecurityClassp, 1,
@@ -904,42 +1117,6 @@ int afsd_InitCM(char **reasonP)
     rx_StartServer(0);
     afsi_log("rx_StartServer");
 
-    /* init user daemon, and other packages */
-    cm_InitUser();
-
-    cm_InitACLCache(2*stats);
-
-    cm_InitConn();
-
-    cm_InitCell();
-        
-    cm_InitServer();
-        
-    cm_InitVolume();
-
-    cm_InitIoctl();
-        
-    smb_InitIoctl();
-        
-    cm_InitCallback();
-        
-    cm_InitSCache(stats);
-        
-    code = cm_InitDCache(0, cacheBlocks);
-    afsi_log("cm_InitDCache code %x", code);
-    if (code != 0) {
-        *reasonP = "error initializing cache";
-        return -1;
-    }
-
-#ifdef AFS_AFSDB_ENV
-#if !defined(_WIN32_WINNT) || (_WIN32_WINNT < 0x0500)
-    if (cm_InitDNS(cm_dnsEnabled) == -1)
-        cm_dnsEnabled = 0;  /* init failed, so deactivate */
-    afsi_log("cm_InitDNS %d", cm_dnsEnabled);
-#endif
-#endif
-
     code = cm_GetRootCellName(rootCellName);
     afsi_log("cm_GetRootCellName code %d, cm_freelanceEnabled= %d, rcn= %s", 
               code, cm_freelanceEnabled, (code ? "<none>" : rootCellName));
@@ -949,13 +1126,13 @@ int afsd_InitCM(char **reasonP)
         return -1;
     }   
     else if (cm_freelanceEnabled)
-        cm_rootCellp = NULL;
+        cm_data.rootCellp = NULL;
 
     if (code == 0 && !cm_freelanceEnabled) 
     {
-        cm_rootCellp = cm_GetCell(rootCellName, CM_FLAG_CREATE);
-        afsi_log("cm_GetCell addr %x", (int)cm_rootCellp);
-        if (cm_rootCellp == NULL) 
+        cm_data.rootCellp = cm_GetCell(rootCellName, CM_FLAG_CREATE);
+        afsi_log("cm_GetCell addr %x", (int)cm_data.rootCellp);
+        if (cm_data.rootCellp == NULL) 
         {
             *reasonP = "can't find root cell in afsdcell.ini";
             return -1;
@@ -966,63 +1143,86 @@ int afsd_InitCM(char **reasonP)
     if (cm_freelanceEnabled)
         cm_InitFreelance();
 #endif
+
+    /* Initialize the RPC server for session keys */
+    RpcInit();
+
+#if !defined(DJGPP)
+    afsd_InitServerPreferences();
+#endif
     return 0;
 }
 
 int afsd_InitDaemons(char **reasonP)
 {
-	long code;
-	cm_req_t req;
+    long code;
+    cm_req_t req;
 
-	cm_InitReq(&req);
+    cm_InitReq(&req);
 
-	/* this should really be in an init daemon from here on down */
+    /* this should really be in an init daemon from here on down */
 
     if (!cm_freelanceEnabled) {
-		osi_Log0(afsd_logp, "Loading Root Volume from cell");
-        code = cm_GetVolumeByName(cm_rootCellp, cm_rootVolumeName, cm_rootUserp,
-                                  &req, CM_FLAG_CREATE, &cm_rootVolumep);
+        osi_Log0(afsd_logp, "Loading Root Volume from cell");
+        code = cm_GetVolumeByName(cm_data.rootCellp, cm_rootVolumeName, cm_rootUserp,
+                                  &req, CM_FLAG_CREATE, &cm_data.rootVolumep);
         afsi_log("cm_GetVolumeByName code %x root vol %x", code,
-                 (code ? (cm_volume_t *)-1 : cm_rootVolumep));
+                 (code ? (cm_volume_t *)-1 : cm_data.rootVolumep));
         if (code != 0) {
             *reasonP = "can't find root volume in root cell";
             return -1;
         }
     }
 
-	/* compute the root fid */
-	if (!cm_freelanceEnabled) {
-        cm_rootFid.cell = cm_rootCellp->cellID;
-        cm_rootFid.volume = cm_GetROVolumeID(cm_rootVolumep);
-        cm_rootFid.vnode = 1;
-        cm_rootFid.unique = 1;
-	}
-	else
-        cm_FakeRootFid(&cm_rootFid);
+    /* compute the root fid */
+    if (!cm_freelanceEnabled) {
+        cm_data.rootFid.cell = cm_data.rootCellp->cellID;
+        cm_data.rootFid.volume = cm_GetROVolumeID(cm_data.rootVolumep);
+        cm_data.rootFid.vnode = 1;
+        cm_data.rootFid.unique = 1;
+    }
+    else
+        cm_FakeRootFid(&cm_data.rootFid);
         
-    code = cm_GetSCache(&cm_rootFid, &cm_rootSCachep, cm_rootUserp, &req);
-	afsi_log("cm_GetSCache code %x scache %x", code,
-             (code ? (cm_scache_t *)-1 : cm_rootSCachep));
-	if (code != 0) {
-		*reasonP = "unknown error";
-		return -1;
-	}
+    code = cm_GetSCache(&cm_data.rootFid, &cm_data.rootSCachep, cm_rootUserp, &req);
+    afsi_log("cm_GetSCache code %x scache %x", code,
+             (code ? (cm_scache_t *)-1 : cm_data.rootSCachep));
+    if (code != 0) {
+        *reasonP = "unknown error";
+        return -1;
+    }
 
-	cm_InitDaemon(numBkgD);
-	afsi_log("cm_InitDaemon");
+    cm_InitDaemon(numBkgD);
+    afsi_log("cm_InitDaemon complete");
 
-	return 0;
+    return 0;
 }
 
 int afsd_InitSMB(char **reasonP, void *aMBfunc)
 {
-	/* Do this last so that we don't handle requests before init is done.
+    HKEY parmKey;
+    DWORD dummyLen;
+    DWORD dwValue;
+    DWORD code;
+
+    code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSREG_CLT_OPENAFS_SUBKEY,
+                         0, KEY_QUERY_VALUE, &parmKey);
+    if (code == ERROR_SUCCESS) {
+        dummyLen = sizeof(DWORD);
+        code = RegQueryValueEx(parmKey, "StoreAnsiFilenames", NULL, NULL,
+                                (BYTE *) &dwValue, &dummyLen);
+        if (code == ERROR_SUCCESS)
+            smb_StoreAnsiFilenames = dwValue ? 1 : 0;
+        RegCloseKey (parmKey);
+    }
+
+    /* Do this last so that we don't handle requests before init is done.
      * Here we initialize the SMB listener.
      */
     smb_Init(afsd_logp, cm_NetbiosName, smb_UseV3, LANadapter, numSvThreads, aMBfunc);
-    afsi_log("smb_Init");
+    afsi_log("smb_Init complete");
 
-	return 0;
+    return 0;
 }
 
 #ifdef ReadOnly
@@ -1187,12 +1387,102 @@ static DWORD *afsd_crtDbgBreakCurrent = NULL;
 static DWORD afsd_crtDbgBreaks[256];
 #endif
 
+static EFaultRepRetVal (WINAPI *pReportFault)(LPEXCEPTION_POINTERS pep, DWORD dwMode) = NULL;
+static BOOL (WINAPI *pMiniDumpWriteDump)(HANDLE hProcess,DWORD ProcessId,HANDLE hFile,
+                                  MINIDUMP_TYPE DumpType,
+                                  PMINIDUMP_EXCEPTION_INFORMATION ExceptionParam,
+                                  PMINIDUMP_USER_STREAM_INFORMATION UserStreamParam,
+                                  PMINIDUMP_CALLBACK_INFORMATION CallbackParam) = NULL;
+
+
+static HANDLE
+OpenDumpFile(void)
+{
+    char wd[256];
+    DWORD code;
+
+    code = GetEnvironmentVariable("TEMP", wd, sizeof(wd));
+    if ( code == 0 || code > sizeof(wd) )
+    {
+        if (!GetWindowsDirectory(wd, sizeof(wd)))
+            return NULL;
+    }
+    StringCbCatA(wd, sizeof(wd), "\\afsd.dmp");
+    return CreateFile( wd, GENERIC_WRITE, FILE_SHARE_READ, NULL,
+                            CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, NULL);
+}
+
+void 
+GenerateMiniDump(PEXCEPTION_POINTERS ep)
+{
+    if (ep == NULL) 
+    {
+        // Generate exception to get proper context in dump
+        __try 
+        {
+            RaiseException(DBG_CONTINUE, 0, 0, NULL);
+        } 
+        __except(GenerateMiniDump(GetExceptionInformation()), EXCEPTION_CONTINUE_EXECUTION) 
+        {
+        }
+    } 
+    else
+    {
+        MINIDUMP_EXCEPTION_INFORMATION eInfo;
+        HANDLE hFile = NULL;
+        HMODULE hDbgHelp = NULL;
+
+        hDbgHelp = LoadLibrary("Dbghelp.dll");
+        if ( hDbgHelp == NULL )
+            return;
+
+        (FARPROC) pMiniDumpWriteDump = GetProcAddress(hDbgHelp, "MiniDumpWriteDump");
+        if ( pMiniDumpWriteDump == NULL ) {
+            FreeLibrary(hDbgHelp);
+            return;
+        }
+
+        hFile = OpenDumpFile();
+
+        if ( hFile ) {
+            HKEY parmKey;
+            DWORD dummyLen;
+            DWORD dwValue;
+            DWORD code;
+            DWORD dwMiniDumpType = MiniDumpNormal;
+
+            code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, AFSREG_CLT_OPENAFS_SUBKEY,
+                                 0, KEY_QUERY_VALUE, &parmKey);
+            if (code == ERROR_SUCCESS) {
+                dummyLen = sizeof(DWORD);
+                code = RegQueryValueEx(parmKey, "MiniDumpType", NULL, NULL,
+                                        (BYTE *) &dwValue, &dummyLen);
+                if (code == ERROR_SUCCESS)
+                    dwMiniDumpType = dwValue ? 1 : 0;
+                RegCloseKey (parmKey);
+            }
+
+            eInfo.ThreadId = GetCurrentThreadId();
+            eInfo.ExceptionPointers = ep;
+            eInfo.ClientPointers = FALSE;
+
+            pMiniDumpWriteDump( GetCurrentProcess(), GetCurrentProcessId(),
+                                hFile, dwMiniDumpType, ep ? &eInfo : NULL,
+                                NULL, NULL);
+
+            CloseHandle(hFile);
+        }
+        FreeLibrary(hDbgHelp);
+    }
+}
+
 LONG __stdcall afsd_ExceptionFilter(EXCEPTION_POINTERS *ep)
 {
     CONTEXT context;
 #ifdef _DEBUG  
     BOOL allocRequestBrk = FALSE;
 #endif 
+    HMODULE hLib = NULL;
   
     afsi_log("UnhandledException : code : 0x%x, address: 0x%x\n", 
              ep->ExceptionRecord->ExceptionCode, 
@@ -1211,7 +1501,17 @@ LONG __stdcall afsd_ExceptionFilter(EXCEPTION_POINTERS *ep)
     context = *ep->ContextRecord;
 	   
     afsd_printStack(GetCurrentThread(), &context);
-	   
+
+    GenerateMiniDump(ep);
+
+    hLib = LoadLibrary("Faultrep.dll");
+    if ( hLib ) {
+        (FARPROC) pReportFault = GetProcAddress(hLib, "ReportFault");
+        if ( pReportFault )
+            pReportFault(ep, 0);
+        FreeLibrary(hLib);
+    }
+
     if (ep->ExceptionRecord->ExceptionCode == EXCEPTION_BREAKPOINT)
     {
         afsi_log("\nEXCEPTION_BREAKPOINT - continue execution ...\n");
