@@ -29,7 +29,7 @@ extern int afs_shuttingdown;
  * Macros to uniquely identify the AFS vfs struct
  */
 #define	AFS_VFSMAGIC		0x1234
-#if    defined(AFS_SUN_ENV) || defined(AFS_HPUX90_ENV) || defined(AFS_LINUX20_ENV)
+#if defined(AFS_SUN5_ENV) || defined(AFS_HPUX90_ENV) || defined(AFS_LINUX20_ENV)
 #define	AFS_VFSFSID		99
 #else
 #if defined(AFS_SGI_ENV)
@@ -43,7 +43,7 @@ extern int afs_shuttingdown;
 #if     defined(AFS_HPUX102_ENV)
 #define AFS_FLOCK       k_flock
 #else
-#if     defined(AFS_SUN56_ENV) || (defined(AFS_LINUX24_ENV) && !defined(AFS_PPC64_LINUX26_ENV) && !defined(AFS_AMD64_LINUX26_ENV) && !defined(AFS_IA64_LINUX26_ENV))
+#if     defined(AFS_SUN56_ENV) || (defined(AFS_LINUX24_ENV) && !(defined(AFS_LINUX26_ENV) && defined(AFS_LINUX_64BIT_KERNEL)))
 #define AFS_FLOCK       flock64
 #else
 #define AFS_FLOCK       flock
@@ -551,7 +551,7 @@ struct SimpleLocks {
 #define VPageCleaning 0x2	/* Solaris - Cache Trunc Daemon sez keep out */
 
 #define	CPSIZE	    2
-#if defined(AFS_XBSD_ENV)
+#if defined(AFS_XBSD_ENV) || defined(AFS_DARWIN_ENV)
 #define vrefCount   v->v_usecount
 #else
 #define vrefCount   v.v_count
@@ -562,19 +562,13 @@ struct SimpleLocks {
 #define VREFCOUNT_SET(v, c)	atomic_set(&((vnode_t *) v)->v_count, c)
 #define VREFCOUNT_DEC(v)	atomic_dec(&((vnode_t *) v)->v_count)
 #define VREFCOUNT_INC(v)	atomic_inc(&((vnode_t *) v)->v_count)
-#define DLOCK()      spin_lock(&dcache_lock)
-#define DUNLOCK()    spin_unlock(&dcache_lock)
-#define DGET(d)      dget_locked(d)
-#define DCOUNT(d)    atomic_read(&(d)->d_count)
 #else
 #define VREFCOUNT(v)		((v)->vrefCount)
 #define VREFCOUNT_SET(v, c)	(v)->vrefCount = c;
 #define VREFCOUNT_DEC(v)	(v)->vrefCount--;
 #define VREFCOUNT_INC(v)	(v)->vrefCount++;
-#define DLOCK()
-#define DUNLOCK()
-#define DGET(d)      dget(d)
-#define DCOUNT(d)    ((d)->d_count)
+#define d_unhash(d) list_empty(&(d)->d_hash)
+#define dget_locked(d) dget(d)
 #endif
 
 #define	AFS_MAXDV   0x7fffffff	/* largest dataversion number */
@@ -586,21 +580,7 @@ struct SimpleLocks {
 
 extern afs_int32 vmPageHog;	/* counter for # of vnodes which are page hogs. */
 
-/*
- * Fast map from vcache to dcache
- */
-struct vtodc {
-    struct dcache *dc;
-    afs_uint32 stamp;
-    struct osi_file *f;
-    afs_offs_t minLoc;		/* smallest offset into dc. */
-    afs_offs_t len;		/* largest offset into dc. */
-};
-
-extern afs_uint32 afs_stampValue;	/* stamp for pair's usage */
-#define	MakeStamp()	(++afs_stampValue)
-
-#if defined(AFS_XBSD_ENV)
+#if defined(AFS_XBSD_ENV) || defined(AFS_DARWIN_ENV)
 #define VTOAFS(v) ((struct vcache *)(v)->v_data)
 #define AFSTOV(vc) ((vc)->v)
 #else
@@ -618,7 +598,7 @@ extern afs_uint32 afs_stampValue;	/* stamp for pair's usage */
  * !(avc->nextfree) && !avc->vlruq.next => (FreeVCList == avc->nextfree)
  */
 struct vcache {
-#if defined(AFS_XBSD_ENV)
+#if defined(AFS_XBSD_ENV)||defined(AFS_DARWIN_ENV)
     struct vnode *v;
 #else
     struct vnode v;		/* Has reference count in v.v_count */
@@ -626,6 +606,7 @@ struct vcache {
     struct afs_q vlruq;		/* lru q next and prev */
     struct vcache *nextfree;	/* next on free list (if free) */
     struct vcache *hnext;	/* Hash next */
+    struct vcache *vhnext; /* vol hash next */
     struct VenusFid fid;
     struct mstat {
 	afs_size_t Length;
@@ -647,7 +628,7 @@ struct vcache {
     krwlock_t rwlock;
     struct cred *credp;
 #endif
-#if defined(AFS_SUN_ENV) || defined(AFS_ALPHA_ENV) || defined(AFS_DARWIN_ENV) || defined(AFS_FBSD_ENV)
+#ifdef AFS_BOZONLOCK_ENV
     afs_bozoLock_t pvnLock;	/* see locks.x */
 #endif
 #ifdef	AFS_AIX32_ENV
@@ -695,14 +676,8 @@ struct vcache {
 #if	defined(AFS_SUN5_ENV)
     afs_uint32 vstates;		/* vstate bits */
 #endif				/* defined(AFS_SUN5_ENV) */
-    struct vtodc quick;
-    afs_uint32 symhintstamp;
-    union {
-	struct vcache *symhint;
-	struct dcache *dchint;
-    } h1;
+    struct dcache *dchint;
 #ifdef AFS_LINUX22_ENV
-    u_short flushcnt;		/* Number of flushes which haven't released yet. */
     u_short mapcnt;		/* Number of mappings of this file. */
 #endif
 #if defined(AFS_SGI_ENV)
@@ -729,9 +704,6 @@ struct vcache {
 #endif
 };
 
-#define afs_symhint_inval(avc)
-
-
 #define	DONT_CHECK_MODE_BITS	0
 #define	CHECK_MODE_BITS		1
 #define CMB_ALLOW_EXEC_AS_READ  2	/* For the NFS xlator */
@@ -791,6 +763,13 @@ struct vcxstat {
     afs_uint32 states;
 };
 
+struct vcxstat2 {
+    afs_int32 callerAccess;
+    afs_int32 cbExpires;
+    afs_int32 anyAccess;
+    char mvstat;
+};
+
 struct sbstruct {
     int sb_thisfile;
     int sb_default;
@@ -835,6 +814,9 @@ struct cm_initparams {
 
 #define	NULLIDX	    (-1)	/* null index definition */
 /* struct dcache states bits */
+#define DRO         1
+#define DBackup     2
+#define DRW         4
 #define	DWriting    8		/* file being written (used for cache validation) */
 
 /* dcache data flags */
@@ -911,6 +893,23 @@ struct afs_fheader {
 #endif
 #endif
 
+struct buffer {
+  afs_int32 fid;              /* is adc->index, the cache file number */
+  afs_inode_t inode;          /* is adc->f.inode, the inode number of the cac\
+				 he file */
+  afs_int32 page;
+  afs_int32 accesstime;
+  struct buffer *hashNext;
+  char *data;
+  char lockers;
+  char dirty;
+  char hashIndex;
+#if AFS_USEBUFFERS
+  struct buf *bufp;
+#endif
+  afs_rwlock_t lock;          /* the lock for this structure */
+};
+
 /* kept on disk and in dcache entries */
 struct fcache {
     struct VenusFid fid;	/* Fid for this file */
@@ -942,8 +941,7 @@ struct dcache {
     char dflags;		/* Data flags */
     char mflags;		/* Meta flags */
     struct fcache f;		/* disk image */
-    afs_int32 stamp;		/* used with vtodc struct for hints */
-
+    afs_int32 bucket;           /* which bucket these dcache entries are in */
     /*
      * Locking rules:
      *
@@ -962,10 +960,6 @@ struct dcache {
      * Note that dcache.lock(W) gives you the right to update mflags,
      * as dcache.mflock(W) can only be held with dcache.lock(R).
      *
-     * dcache.stamp is protected by the associated vcache lock, because
-     * it's only purpose is to establish correspondence between vcache
-     * and dcache entries.
-     *
      * dcache.index, dcache.f.fid, dcache.f.chunk and dcache.f.inode are
      * write-protected by afs_xdcache and read-protected by refCount.
      * Once an entry is referenced, these values cannot change, and if
@@ -975,8 +969,14 @@ struct dcache {
      * ensuring noone else has a refCount on it).
      */
 };
-/* this is obsolete and should be removed */
-#define ihint stamp
+
+/* afs_memcache.c */
+struct memCacheEntry {
+  int size;                   /* # of valid bytes in this entry */
+  int dataSize;               /* size of allocated data area */
+  afs_lock_t afs_memLock;
+  char *data;                 /* bytes */
+};
 
 /* macro to mark a dcache entry as bad */
 #define ZapDCE(x) \
@@ -1037,6 +1037,8 @@ struct dcache {
 /* don't hash on the cell, our callback-breaking code sometimes fails to compute
     the cell correctly, and only scans one hash bucket */
 #define	VCHash(fid)	(((fid)->Fid.Volume + (fid)->Fid.Vnode) & (VCSIZE-1))
+/* Hash only on volume to speed up volume callbacks. */
+#define VCHashV(fid) ((fid)->Fid.Volume & (VCSIZE-1))
 
 extern struct dcache **afs_indexTable;	/*Pointers to in-memory dcache entries */
 extern afs_int32 *afs_indexUnique;	/*dcache entry Fid.Unique */
@@ -1046,6 +1048,7 @@ extern afs_int32 afs_cacheFiles;	/*Size of afs_indexTable */
 extern afs_int32 afs_cacheBlocks;	/*1K blocks in cache */
 extern afs_int32 afs_cacheStats;	/*Stat entries in cache */
 extern struct vcache *afs_vhashT[VCSIZE];	/*Stat cache hash table */
+extern struct vcache *afs_vhashTV[VCSIZE]; /* cache hash table on volume */
 extern afs_int32 afs_initState;	/*Initialization state */
 extern afs_int32 afs_termState;	/* Termination state */
 extern struct VenusFid afs_rootFid;	/*Root for whole file system */
@@ -1194,29 +1197,6 @@ struct afs_fakestat_state {
 
 extern int afs_fakestat_enable;
 
-struct buffer {
-    struct dcache *fid;
-    afs_int32 page;
-    afs_int32 accesstime;
-    struct buffer *hashNext;
-    char *data;
-    char lockers;
-    char dirty;
-    char hashIndex;
-#if AFS_USEBUFFERS
-    struct buf *bufp;
-#endif
-    afs_rwlock_t lock;		/* the lock for this structure */
-};
-
-/* afs_memcache.c */
-struct memCacheEntry {
-    int size;			/* # of valid bytes in this entry */
-    int dataSize;		/* size of allocated data area */
-    afs_lock_t afs_memLock;
-    char *data;			/* bytes */
-};
-
 /* First 32 bits of capabilities */
 #define CAPABILITY_ERRORTRANS (1<<0)
 
