@@ -14,7 +14,7 @@
 #include "afs/param.h"
 
 RCSID
-    ("$Header: /cvs/openafs/src/afs/afs_cell.c,v 1.29 2003/07/15 23:14:11 shadow Exp $");
+    ("$Header: /cvs/openafs/src/afs/afs_cell.c,v 1.30.2.1 2005/04/06 05:09:09 shadow Exp $");
 
 #include "afs/stds.h"
 #include "afs/sysincludes.h"	/* Standard vendor system headers */
@@ -819,6 +819,7 @@ afs_CellInit()
 {
     RWLOCK_INIT(&afs_xcell, "afs_xcell");
 #ifdef AFS_AFSDB_ENV
+    RWLOCK_INIT(&afsdb_client_lock, "afsdb_client_lock");
     RWLOCK_INIT(&afsdb_req.lock, "afsdb_req.lock");
 #endif
     QInit(&CellLRU);
@@ -843,6 +844,18 @@ shutdown_cell()
 	afs_osi_Free(tc, sizeof(struct cell));
     }
     QInit(&CellLRU);
+
+{
+    struct cell_name *cn = afs_cellname_head;
+
+    while (cn) {
+	struct cell_name *next = cn->next;
+
+	afs_osi_FreeStr(cn->cellname);
+	afs_osi_Free(cn, sizeof(struct cell_name));
+	cn = next;
+    }
+}
 }
 
 void
