@@ -21,6 +21,7 @@ int newVLDB;			/* Compatibility flag */
 #endif
 static int newVLDB = 1;
 
+
 #ifndef AFS_PTHREAD_ENV
 #define USUAL_PRIORITY (LWP_MAX_PRIORITY - 2)
 
@@ -35,11 +36,21 @@ static int newVLDB = 1;
    fsync.c
    File server synchronization with external volume utilities.
  */
+
+/* This controls the size of an fd_set; it must be defined early before
+ * the system headers define that type and the macros that operate on it.
+ * Its value should be as large as the maximum file descriptor limit we
+ * are likely to run into on any platform.  Right now, that is 65536
+ * which is the default hard fd limit on Solaris 9 */
+#ifndef _WIN32
+#define FD_SETSIZE 65536
+#endif
+
 #include <afsconfig.h>
 #include <afs/param.h>
 
 RCSID
-    ("$Header: /cvs/openafs/src/vol/fssync.c,v 1.21 2003/11/15 04:59:14 shadow Exp $");
+    ("$Header: /cvs/openafs/src/vol/fssync.c,v 1.26.2.3 2005/04/27 01:37:09 shadow Exp $");
 
 #include <sys/types.h>
 #include <stdio.h>
@@ -90,12 +101,11 @@ RCSID
 #endif
 #define osi_Assert(e) (void)(e)
 
-extern int LogLevel;		/* Vice loglevel */
 int (*V_BreakVolumeCallbacks) ();
 
 #define MAXHANDLERS	4	/* Up to 4 clients; must be at least 2, so that
 				 * move = dump+restore can run on single server */
-#define MAXOFFLINEVOLUMES 30	/* This needs to be as big as the maximum
+#define MAXOFFLINEVOLUMES 128	/* This needs to be as big as the maximum
 				 * number that would be offline for 1 operation.
 				 * Current winner is salvage, which needs all
 				 * cloned read-only copies offline when salvaging
@@ -123,9 +133,7 @@ struct command {
     char partName[16];		/* partition name, e.g. /vicepa */
 };
 
-
 /* Forward declarations */
-static int getport();
 static void FSYNC_sync();
 static void FSYNC_newconnection();
 static void FSYNC_com();
@@ -140,6 +148,8 @@ static int FindHandler_r();
 static int RemoveHandler();
 static void GetHandler(fd_set * fdsetp, int *maxfdp);
 
+extern int LogLevel;
+
 /*
  * This lock controls access to the handler array. The overhead
  * is minimal in non-preemptive environments.
@@ -159,18 +169,10 @@ FSYNC_clientInit(void)
 	FS_sd = getport(&addr);
 	if (connect(FS_sd, (struct sockaddr *)&addr, sizeof(addr)) >= 0)
 	    return 1;
-#if defined(AFS_SGI_ENV)
-	/* down with worthless error messages! */
-	if (!*timeout) {
-	    perror("FSYNC_clientInit failed (after many retries)");
-	    break;
-	}
-#else
 	if (!*timeout)
 	    break;
 	if (!(*timeout & 1))
-	    perror("FSYNC_clientInit temporary failure (will retry)");
-#endif
+	    Log(0, ("FSYNC_clientInit temporary failure (will retry)"));
 	FSYNC_clientFinis();
 	sleep(*timeout++);
     }
@@ -187,7 +189,6 @@ FSYNC_clientFinis(void)
     close(FS_sd);
 #endif
     FS_sd = -1;
-    Lock_Destroy(&FSYNC_handler_lock);
 }
 
 int
@@ -204,7 +205,7 @@ FSYNC_askfs(VolumeId volume, char *partName, int com, int reason)
     else
 	command.partName[0] = 0;
     assert(FS_sd != -1);
-    VFSYNC_LOCK
+    VFSYNC_LOCK;
 #ifdef AFS_NT40_ENV
     if (send(FS_sd, (char *)&command, sizeof(command), 0) != sizeof(command)) {
 	printf("FSYNC_askfs: write to file server failed\n");
@@ -237,9 +238,8 @@ FSYNC_askfs(VolumeId volume, char *partName, int com, int reason)
 	    ("FSYNC_askfs: negative response from file server; volume %u, command %d\n",
 	     command.volume, (int)command.command);
     }
- done:
-    VFSYNC_UNLOCK
-
+  done:
+    VFSYNC_UNLOCK;
     return response;
 }
 
@@ -277,6 +277,8 @@ getport(struct sockaddr_in *addr)
     return sd;
 }
 
+static fd_set FSYNC_readfds;
+
 static void
 FSYNC_sync()
 {
@@ -332,18 +334,17 @@ FSYNC_sync()
     InitHandler();
     AcceptOn();
     for (;;) {
-	fd_set readfds;
 	int maxfd;
-	GetHandler(&readfds, &maxfd);
+	GetHandler(&FSYNC_readfds, &maxfd);
 	/* Note: check for >= 1 below is essential since IOMGR_select
 	 * doesn't have exactly same semantics as select.
 	 */
 #ifdef AFS_PTHREAD_ENV
-	if (select(maxfd + 1, &readfds, NULL, NULL, NULL) >= 1)
+	if (select(maxfd + 1, &FSYNC_readfds, NULL, NULL, NULL) >= 1)
 #else /* AFS_PTHREAD_ENV */
-	if (IOMGR_Select(maxfd + 1, &readfds, NULL, NULL, NULL) >= 1)
+	if (IOMGR_Select(maxfd + 1, &FSYNC_readfds, NULL, NULL, NULL) >= 1)
 #endif /* AFS_PTHREAD_ENV */
-	    CallHandler(&readfds);
+	    CallHandler(&FSYNC_readfds);
     }
 }
 
@@ -395,7 +396,9 @@ FSYNC_com(int fd)
 	FSYNC_Drop(fd);
 	return;
     }
-    VATTACH_LOCK VOL_LOCK volumes = OfflineVolumes[FindHandler(fd)];
+    VATTACH_LOCK;
+    VOL_LOCK;
+    volumes = OfflineVolumes[FindHandler(fd)];
     for (v = 0, i = 0; i < MAXOFFLINEVOLUMES; i++) {
 	if (volumes[i].volumeID == command.volume
 	    && strcmp(volumes[i].partName, command.partName) == 0) {
@@ -557,28 +560,35 @@ defect #2080 for details.
 	if (V_BreakVolumeCallbacks) {
 	    Log("fssync: volume %u moved to %x; breaking all call backs\n",
 		command.volume, command.reason);
-	    VOL_UNLOCK VATTACH_UNLOCK(*V_BreakVolumeCallbacks) (command.
-								volume);
-	VATTACH_LOCK VOL_LOCK}
+	    VOL_UNLOCK;
+	    VATTACH_UNLOCK;
+	    (*V_BreakVolumeCallbacks) (command.volume);
+	    VATTACH_LOCK;
+	    VOL_LOCK;
+	}
 	break;
     case FSYNC_RESTOREVOLUME:
 	/* if the volume is being restored, break all callbacks on it */
 	if (V_BreakVolumeCallbacks) {
 	    Log("fssync: volume %u restored; breaking all call backs\n",
 		command.volume);
-	    VOL_UNLOCK VATTACH_UNLOCK(*V_BreakVolumeCallbacks) (command.
-								volume);
-	VATTACH_LOCK VOL_LOCK}
+	    VOL_UNLOCK;
+	    VATTACH_UNLOCK;
+	    (*V_BreakVolumeCallbacks) (command.volume);
+	    VATTACH_LOCK;
+	    VOL_LOCK;
+	}
 	break;
     default:
 	rc = FSYNC_DENIED;
 	break;
     }
-    VOL_UNLOCK VATTACH_UNLOCK
+    VOL_UNLOCK;
+    VATTACH_UNLOCK;
 #ifdef AFS_NT40_ENV
-      (void) send(fd, &rc, 1, 0);
+    (void)send(fd, &rc, 1, 0);
 #else
-      (void) write(fd, &rc, 1);
+    (void)write(fd, &rc, 1);
 #endif
 }
 
@@ -590,7 +600,9 @@ FSYNC_Drop(int fd)
     Error error;
     char tvolName[VMAXPATHLEN];
 
-    VATTACH_LOCK VOL_LOCK p = OfflineVolumes[FindHandler(fd)];
+    VATTACH_LOCK;
+    VOL_LOCK;
+    p = OfflineVolumes[FindHandler(fd)];
     for (i = 0; i < MAXOFFLINEVOLUMES; i++) {
 	if (p[i].volumeID) {
 	    Volume *vp;
@@ -604,7 +616,9 @@ FSYNC_Drop(int fd)
 	    p[i].volumeID = 0;
 	}
     }
-    VOL_UNLOCK VATTACH_UNLOCK RemoveHandler(fd);
+    VOL_UNLOCK;
+    VATTACH_UNLOCK;
+    RemoveHandler(fd);
 #ifdef AFS_NT40_ENV
     closesocket(fd);
 #else
